# -*- coding: utf-8 -*-
import mimetypes
import os
import re
import uuid
from hashlib import md5

import magic
from autoslug import AutoSlugField
from django.conf import settings
from django.core.exceptions import ValidationError, SuspiciousOperation
from django.core.validators import (
    MinValueValidator,
    MaxValueValidator,
    validate_slug,
    FileExtensionValidator,
    MinLengthValidator,
)
from django.db import models
from django.db.models.aggregates import Sum
from django.http import Http404
from django.template.defaultfilters import slugify
from django.urls import reverse
from django.utils import timezone
from django.utils.crypto import get_random_string
from django.utils.deconstruct import deconstructible
from django.utils.html import escape
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from martor.models import MartorField
from multiselectfield import MultiSelectField
from paypal.standard.ipn.models import PayPalIPN
from paypal.standard.models import ST_PP_COMPLETED

from accounts.models import User, Group


def short_url_validate(value):
    if value == "":
        return
    if re.match(r"^[\d]+$", value):
        raise ValidationError(
            _("%(value)s doit être composée au moins d'un caractère alphabétique ou _"),
            params={"value": value},
        )
    if value in ["add"]:
        raise ValidationError(
            _("%(value)s est une valeur réservée"),
            params={"value": value},
        )
    return


def short_url_slug(value, nb=0):
    v = slugify(value + ("-" + str(nb) if nb != 0 else ""))
    try:
        short_url_validate(v)
    except ValidationError:
        return short_url_slug(value, nb + 1)
    return v


class PaypalAccount(models.Model):
    """
    Compte Paypal sur lesquels des fonds peuvent être virés
    """

    email = models.EmailField(unique=True)
    description = models.CharField(max_length=100, verbose_name="Description du compte PayPal")
    no_account = models.BooleanField(default=False, editable=False)
    secure_form = models.BooleanField(default=False)
    cert_id = models.CharField(max_length=32, blank=True)
    cert_path_pub = models.CharField(max_length=100, blank=True)
    cert_path_priv = models.CharField(max_length=100, blank=True)
    cert_end_date = models.DateTimeField(default=timezone.now)

    def secure_available(self):
        return self.secure_form and self.cert_end_date > timezone.now()

    def delete(self, *args, **kwargs):
        if not self.no_account:
            super(PaypalAccount, self).delete(*args, **kwargs)
        else:
            raise SuspiciousOperation("No account can't be deleted")

    def save(self, *args, **kwargs):
        if not self.no_account:
            super(PaypalAccount, self).save(*args, **kwargs)
        else:
            raise SuspiciousOperation("No account can't be edited")

    def __str__(self):
        if self.no_account:
            return "Pas de compte, Paypal sera désactivé pour cet évènement."
        if self.secure_available():
            return "{} – {} (sécurisé {} jusqu'au {:%d/%m/%Y})".format(
                self.email, self.description, self.cert_id, self.cert_end_date
            )
        else:
            return "{} – {} (non sécurisé)".format(self.email, self.description)

    def get_cert_priv(self):
        return os.path.join(settings.PAYPAL_PATH_CERT, self.cert_path_priv)

    def get_cert_pub(self):
        return os.path.join(settings.PAYPAL_PATH_CERT, self.cert_path_pub)

    @classmethod
    def default_paypal_account(cls):
        no_account = cls.objects.filter(no_account=True)
        if no_account.exists():
            return no_account.first()

        no_account = cls(email="no_account@billetterie.inpt.fr", no_account=True)
        super(cls, no_account).save()
        return no_account

    class Meta:
        ordering = ["-no_account"]


class LydiaAccount(models.Model):
    """
    Compte Lydia sur lequel des fonds peuvent être virés
    """

    vendor_token = models.CharField(
        max_length=settings.LYDIA_TOKEN_LENGTH,
        validators=[MinLengthValidator(settings.LYDIA_TOKEN_LENGTH)],
        verbose_name="Token API",
        null=True,
    )
    private_token = models.CharField(
        max_length=settings.LYDIA_TOKEN_LENGTH,
        validators=[MinLengthValidator(settings.LYDIA_TOKEN_LENGTH)],
        verbose_name="Token privé",
        null=True,
    )
    description = models.CharField(max_length=100, verbose_name="Description du compte Lydia")
    no_account = models.BooleanField(default=False, editable=False)

    class Meta:
        ordering = ["-no_account"]
        unique_together = ["vendor_token", "private_token"]

    @classmethod
    def default_lydia_account(cls):
        no_account = cls.objects.filter(no_account=True)
        if no_account.exists():
            return no_account.first()

        no_account = cls(no_account=True, vendor_token="0" * settings.LYDIA_TOKEN_LENGTH)
        super(cls, no_account).save()
        return no_account

    def __str__(self):
        if self.no_account:
            return "Pas de compte, Lydia sera désactivé pour cet évènement."
        else:
            return self.description

    def get_signature(self, param):
        """
        Générer une signature faite à partir des paramètres de la requête et du token privé, nécessaire pour
        certains appels à l'API et pour vérifier les signatures envoyées par Lydia.
        Voir https://homologation.lydia-app.com/doc/api/#signature pour les détails.
        """

        sorted_param = [(str(k), str(param[k])) for k in sorted(param.keys())]

        sig = []
        for key, val in sorted_param:
            sig.append(key + "=" + val)

        encoded_param = "&".join(sig) + "&" + self.private_token

        return md5(encoded_param.encode("utf-8")).hexdigest()

    def signature_is_valid(self, signature, param):
        return signature == self.get_signature(param)


class Event(models.Model):
    """
    Un Event représente un évènement (e.g. le Gala).
    """

    EVENT_PUBLIC = 1
    EVENT_LOGIN = 2
    EVENT_PRIVE = 3
    EVENT_HIDDEN = 4
    EVENT_CHOICES = (
        (EVENT_PUBLIC, "Évènement public"),
        (EVENT_LOGIN, "Évènement réservé aux membres connectés"),
        (EVENT_PRIVE, "Évènement privé (restriction par club)"),
        (EVENT_HIDDEN, "Évènement caché (non listé)"),
    )

    class Meta:
        ordering = ["start_date"]

    name = models.CharField(max_length=32, verbose_name="Nom de l’évènement")
    description = MartorField(
        blank=True,
        verbose_name="Description de l’évènement",
        default="Aucune description de l'évènement n'a été fournie.\n"
        "Pour obtenir plus d'information, contactez l'organisateur de l'évènement.",
    )
    email_message = models.TextField(
        blank=True,
        verbose_name="Message envoyé par email avec les billets",
        default="Vous trouverez ci-joint votre/vos billets :\n"
        "Ceux-ci sont nominatifs et doivent être imprimés "
        "pour garantir l'accès à l'évènement.\n\n"
        "Vous pouvez à tout moment télécharger à nouveau ces billets "
        "dans la rubrique « Mes Places » de la billetterie INPT.",
    )
    other_payment_description = models.TextField(
        blank=True,
        verbose_name="Description des autres moyens de paiement acceptés que PayPal et Lydia",
        default="Aucune information complémentaire sur les moyens de paiement n'est disponible. "
        "Pour obtenir plus d'information, contactez l'organisateur de l'évènement.",
    )
    terms_of_use = models.TextField(
        blank=True,
        verbose_name="Description des conditions d’utilisation de la billetterie",
        default="Pour obtenir une copie des conditions d'utilisation de la billetterie "
        "et des conditions de vente spécifiques à cet évènement, "
        "veuillez contacter l'organisateur de l'évènement.\n\n"
        "Notez que pour réaliser un achat sur la billetterie INPT, "
        "vous devez avoir lu et accepté ces conditions.",
    )
    start_date = models.DateTimeField(verbose_name="Date et heure de début")
    end_date = models.DateTimeField(verbose_name="Date et heure de fin")
    paypal_account = models.ForeignKey(
        PaypalAccount,
        related_name="paypal_account",
        verbose_name="Email du compte PayPal associé à l’évènement",
        on_delete=models.SET(PaypalAccount.default_paypal_account),
    )
    lydia_account = models.ForeignKey(
        LydiaAccount,
        on_delete=models.SET(LydiaAccount.default_lydia_account),
        verbose_name="Compte Lydia associé à l'évènement",
        null=True,
    )
    event_confidentiality = models.IntegerField(
        choices=EVENT_CHOICES, default=EVENT_PUBLIC, verbose_name="Confidentialité de l’évènement"
    )
    limit_godson = models.PositiveIntegerField(
        default=0, verbose_name="Nombre de parrainage autorisé par personne (0 = pas de parrainage)"
    )
    limit_ticket = models.PositiveIntegerField(
        default=0, verbose_name="Nombre de billet maximum pour l'évènement (0 = non limité)"
    )
    email_event = models.EmailField(verbose_name="Email de contact de l'organisation (assistance aux participants)")
    short_url = AutoSlugField(
        null=True,
        default=None,
        unique=True,
        populate_from="name",
        editable=True,
        max_length=32,
        validators=[short_url_validate],
        slugify=short_url_slug,
    )
    # défini si un évènement est archivé
    # un évènement archivé ne possède plus de logs, registration, ticket mais uniquement les statistiques
    archived = models.BooleanField(default=False)
    archived_stats = models.TextField(blank=True, default="")

    def url(self):
        if self.short_url:
            return self.short_url
        return self.id

    def get_access_manager(self, user):
        if not user.is_authenticated:
            return Manager.NO_ACCESS
        if user.is_superuser:
            return Manager.FULL_MANAGER
        user_acl = UserManager.objects.filter(event=self, user=user)
        access = user_acl.first().access if user_acl.exists() else Manager.NO_ACCESS
        for acl in GroupManager.objects.filter(event=self):
            access = max(access, acl.get_access_right(user))
        return access

    def is_manager(self, user):
        return user.is_authenticated and (user.is_superuser or self.get_access_manager(user) > Manager.NO_ACCESS)

    @property
    def places_remaining(self):
        return max(
            0,
            self.limit_ticket
            - sum([ticket.purchased_places - ticket.places_rejected for ticket in self.tickets.all()]),
        )

    def is_remain(self):
        return self.limit_ticket == 0 or self.places_remaining > 0

    @property
    def registrations(self):
        return Registration.objects.filter(ticket__event=self)

    def at_least_one_ticket_accessible(self, user):
        if self.is_manager(user):
            return True
        for t in self.tickets.all():
            if t.is_accessible_for(user):
                return True

    def has_customer(self, user):
        for registration in user.registrations.all():
            if registration.ticket.event == self:
                return True
        return False

    def __str__(self):
        return self.name

    @classmethod
    def get_event_for_request(cls, request, event_id, archived=False):

        # Search event
        if re.match(r"^[0-9]+$", event_id):
            events = cls.objects.filter(id=event_id)
        else:
            events = cls.objects.filter(short_url=event_id)

        if archived is not None:
            events = events.filter(archived=archived)

        if events.count() != 1:
            raise Http404

        event = events.first()
        request.event = event

        if request.user.is_authenticated:
            request.managers_access = request.event.get_access_manager(request.user)
        else:
            request.managers_access = Manager.NO_ACCESS

        return event


class Manager(models.Model):
    """
    Définit la permission de voir/gérer un évènement à un utilisateur ou un groupe
    """

    NO_ACCESS = 0  # pas mis dans les choix disponibles
    FULL_VIEW = 1
    ENTRY_MANAGER = 2
    EVENT_MANAGER = 3
    FULL_MANAGER = 4
    ACCESS_RIGHT = (
        (FULL_VIEW, "Accès en lecture"),
        (ENTRY_MANAGER, "Accès en lecture et gestion des entrées"),
        (EVENT_MANAGER, "Droit de gestion complet sauf permission"),
        (FULL_MANAGER, "Droit de gestion complet"),
    )

    event = models.ForeignKey(Event, on_delete=models.CASCADE)
    access = models.IntegerField(choices=ACCESS_RIGHT, default=FULL_VIEW, verbose_name="Droit de gestion")

    def __str__(self):
        return "{} - {}".format(self.event, self.access)

    class Meta:
        abstract = True


class UserManager(Manager):
    user = models.ForeignKey(User, on_delete=models.CASCADE)

    def __str__(self):
        return "User {} ({}) of {}".format(self.user, self.access, self.event)


class GroupManager(Manager):
    group = models.ForeignKey(Group, null=True, on_delete=models.CASCADE, verbose_name="Groupe")
    bureau_only = models.BooleanField(default=False, verbose_name="Uniquement le bureau")

    def get_access_right(self, user):
        if self.group in user.managed_clubs.all() or ((not self.bureau_only) and self.group in user.groupes.all()):
            return self.access
        return self.NO_ACCESS

    def __str__(self):
        if self.bureau_only:
            return "Bureau {} ({}) of {}".format(self.group, self.access, self.event)
        else:
            return "Club {} ({}) of {}".format(self.group, self.access, self.event)


def path_design(instance, filename):
    filename, file_extension = os.path.splitext(filename)
    random_str = get_random_string(length=12)
    return "protected/billetterie/event_{0}/template/billet_{1}_{2}{3}".format(
        instance.event.pk, instance.id, random_str, file_extension
    )


@deconstructible
class FileValidator(object):
    def __call__(self, value):
        __, ext = os.path.splitext(value.name)
        ext = ext.lower()
        detected_type = magic.from_buffer(value.read(2048), mime=True)
        value.seek(0)
        if ext not in mimetypes.guess_all_extensions(detected_type):
            raise ValidationError("L'extension ne correpond pas au fichier", "invalid_extension")


@deconstructible
class FileSizeValidator(object):
    def __call__(self, value):
        if value.size > settings.MAX_UPLOAD_SIZE:
            raise ValidationError(
                "Le fichier dépasse {} Mo".format(settings.MAX_UPLOAD_SIZE / (1024 * 1024)), "invalid_size"
            )


class Ticket(models.Model):
    """
    Un Ticket est associé à un évènement.
    Il peut y avoir plusieurs tickets pour le même évènement dans le cas
    où plusieurs tarifs sont proposés (e.g. adhérent AE / extérieur).
    """

    SCHOOL_INP = 1
    SCHOOL_N7 = 2
    SCHOOL_ENSAT = 3
    SCHOOL_A7 = 4
    SCHOOL_CHOICES = (
        (SCHOOL_INP, "Étudiants INP (toute école)"),
        (SCHOOL_N7, "Étudiants N7"),
        (SCHOOL_ENSAT, "Étudiants ENSAT"),
        (SCHOOL_A7, "Étudiants A7"),
    )

    YEAR_1A = 1
    YEAR_2A = 2
    YEAR_3A = 3
    YEAR_4A = 4
    YEAR_5APLUS = 5
    YEAR_CHOICES = (
        (YEAR_1A, "Étudiants en première année"),
        (YEAR_2A, "Étudiants en deuxième année"),
        (YEAR_3A, "Étudiants en troisième année"),
        (YEAR_4A, "Diplomés de l'année courante"),
        (YEAR_5APLUS, "Diplomés des années précédentes"),
    )

    primarykey = models.AutoField(primary_key=True)
    event = models.ForeignKey(Event, related_name="tickets", on_delete=models.CASCADE)
    id = models.IntegerField(editable=False)
    name = models.CharField(max_length=32, verbose_name="Nom")
    short_notes = models.CharField(max_length=256, blank=True, verbose_name="Remarques (courtes, liste des billets)")
    long_notes = models.TextField(blank=True, default="", verbose_name="Remarques (longues, page de confirmation)")
    price = models.DecimalField(
        default=0,
        max_digits=6,
        decimal_places=2,
        verbose_name="Prix (0 pour un billet gratuit)",
        validators=[MinValueValidator(0.000, message="Le prix d’un billet doit être positif ou nul.")],
    )
    limit = models.PositiveIntegerField(default=0, verbose_name="Nombre de place disponible (0 = illimité)")
    min_purchase_date = models.DateTimeField(null=True, blank=True, verbose_name="Date début de vente")
    max_purchase_date = models.DateTimeField(null=True, blank=True, verbose_name="Date limite d’achat (si prévente)")
    online_payment = models.BooleanField(default=True, verbose_name="Paiement en ligne possible")
    allowed_external = models.BooleanField(default=False)
    required_school = models.IntegerField(
        choices=SCHOOL_CHOICES, default=SCHOOL_INP, verbose_name="Restriction par école"
    )
    groups = models.ManyToManyField(
        Group, related_name="authorized_groups", verbose_name="Restriction par groupe", blank=True
    )
    year_choice = MultiSelectField(choices=YEAR_CHOICES, blank=True, verbose_name="Restriction par année")
    required_ae = models.BooleanField(default=True, verbose_name="Réserver ce billet aux adhérents AE")
    allowed_godson = models.BooleanField(
        default=False, verbose_name="Ce billet est accessible aux personnes parrainées"
    )
    design = models.ImageField(
        upload_to=path_design,
        null=True,
        blank=True,
        verbose_name="Template du billet",
        validators=[FileValidator(), FileSizeValidator()],
    )
    admin_ticket = models.BooleanField(default=False, verbose_name="Seul un manager peut donner ce billet")
    allow_anonyme = models.BooleanField(
        default=False,
        verbose_name="Vente sur Place sans compte billetterie (certaines restrictions ne pourront pas être vérifiées)",
    )

    class Meta:
        ordering = ["id"]
        unique_together = ["event", "id"]

    @staticmethod
    def next_id(event):

        last_ticket = event.tickets.last()
        if last_ticket:
            return last_ticket.id + 1
        else:
            return 1

    @property
    def purchased_places(self):
        return self.registrations.filter(paid=True).count()

    @property
    def places_rejected(self):
        return self.registrations.filter(opposed=True).count()

    @property
    def places_remaining(self):
        return max(0, self.limit - self.purchased_places + self.places_rejected)

    def is_remain(self):
        return self.event.is_remain() and (self.limit == 0 or self.places_remaining > 0)

    def still_available(self):
        return (not self.max_purchase_date or self.max_purchase_date >= timezone.now()) and (
            not self.min_purchase_date or self.min_purchase_date <= timezone.now()
        )

    def get_year_choice(self):
        return [] if self.year_choice is None else [int(i) for i in self.year_choice]

    def is_accessible_for(self, user, with_admin=False):

        if self.admin_ticket and not with_admin:
            return False

        if self.allowed_external:
            return True

        if self.required_ae and not user.is_adherent_ae:
            return False

        if self.groups.all():
            condition_user_group = False
        else:
            condition_user_group = True
        for g in user.groupes.all():
            if g in self.groups.all():
                condition_user_group = True

        # yearp0 correspond à l'année de sortie des 0A à la date de début de l'évènement
        yearp0 = self.event.start_date.year + 3
        if self.event.start_date.month in [9, 10, 11, 12]:
            yearp0 += 1

        condition_user_year = True
        if self.get_year_choice():
            condition_user_year = (yearp0 - user.promo in self.get_year_choice()) or (
                yearp0 - user.promo >= Ticket.YEAR_5APLUS and Ticket.YEAR_5APLUS in self.get_year_choice()
            )

        condition_user_school = True
        if self.required_school == Ticket.SCHOOL_INP:
            condition_user_school = user.ecole in ["n7", "ensat", "a7"]

        if self.required_school == Ticket.SCHOOL_N7:
            condition_user_school = user.ecole == "n7"
        if self.required_school == Ticket.SCHOOL_ENSAT:
            condition_user_school = user.ecole == "ensat"
        if self.required_school == Ticket.SCHOOL_A7:
            condition_user_school = user.ecole == "a7"

        return condition_user_group and condition_user_year and condition_user_school

    def __str__(self):
        return "{} – {}".format(self.event.name, self.name)


class Option(models.Model):
    name = models.CharField(max_length=32, verbose_name="Nom")
    price = models.DecimalField(
        max_digits=6,
        decimal_places=2,
        verbose_name="Prix",
        validators=[MinValueValidator(0.000, message="Le prix d’une option doit être positif ou nul.")],
    )
    description = models.TextField(blank=True, verbose_name="Description de l’option")
    group_name = models.CharField(
        blank=True,
        null=True,
        max_length=64,
        verbose_name="Nom du groupe d’option (optionnel, nécessaire "
        "que dans le cas d’options mutuellement exclusives)",
    )
    group = models.IntegerField(
        blank=True,
        null=True,
        default=None,
        verbose_name="Numéro du groupe (optionnel, nécessaire que dans " "les cas d’options mutuellement exclusives)",
    )
    ticket = models.ForeignKey(Ticket, related_name="options", on_delete=models.CASCADE)
    admin_only = models.BooleanField(
        default=False, verbose_name="Option uniquement visible par les administrateurs de l'événement."
    )
    mandatory = models.BooleanField(
        default=False, verbose_name="Option obligatoire " "(l'une des options doit être gratuite et disponible)"
    )

    @staticmethod
    def get_new_ID_group(ticket):
        option_gid = set([o.group for o in Option.objects.filter(ticket=ticket) if o.group])
        new_id = 1
        while new_id in option_gid:
            new_id += 1
        return new_id

    def search_name(self):
        if self.group_name:
            return "{} - {} - {} (+ {}€)".format(self.ticket.name, self.group_name, self.name, self.price)
        else:
            return "{} - {} (+ {}€)".format(self.ticket.name, self.name, self.price)

    def __str__(self):
        return "{} (+ {}€)".format(self.name, self.price)


class Registration(models.Model):
    """
    L’objet Registration représente un achat effectué par l’utilisateur owner.
    Le champ payee indique que le billet a été acheté pour une connaissance ;
    il contient alors le nom de celle-ci. Le reste du temps, le champ est nul.
    """

    PAYMENT_UNKNOW = 0
    PAYMENT_PAYPAL = 1
    PAYMENT_CASH = 2
    PAYMENT_CHECK = 3
    PAYMENT_CCT = 4  # Credit Card Terminal
    PAYMENT_BANK_TRANSFER = 5
    PAYMENT_LYDIA = 6
    PAYMENT_METHODS_DICT = {
        PAYMENT_UNKNOW: "Inconnu",
        PAYMENT_PAYPAL: "Paypal",
        PAYMENT_CASH: "Liquide",
        PAYMENT_CHECK: "Chèque",
        PAYMENT_CCT: "CB",
        PAYMENT_BANK_TRANSFER: "Virement",
        PAYMENT_LYDIA: "Lydia",
    }
    PAYMENT_METHODS = list(PAYMENT_METHODS_DICT.items())

    ALLOWED_ADMIN_PAYMENT = (
        PAYMENT_CASH,
        PAYMENT_CHECK,
        PAYMENT_CCT,
        PAYMENT_BANK_TRANSFER,
    )

    STATE_CANNOT_PAID = 0
    STATE_CAN_PAID = 1
    STATE_CAN_ENTER = 2
    STATE_INFO = (
        (STATE_CANNOT_PAID, "Manque des informations avant le paiement"),
        (STATE_CAN_PAID, "Manque des informations avant l'évènement"),
        (STATE_CAN_ENTER, "Toutes les informations ont été fournies"),
    )

    OPPOSED_UNKNOW = 0
    OPPOSED_BY_ADMIN = 1
    OPPOSED_BY_PAYPAL = 2
    REFUND_BY_ADMIN = 3
    OPPOSED_REASON = (
        (OPPOSED_UNKNOW, "Mise en opposition pour une raison inconnue"),
        (OPPOSED_BY_ADMIN, "Mise en opposition par un administrateur de l'évènement"),
        (OPPOSED_BY_PAYPAL, "Mise en opposition lors du paiement par PayPal"),
        (REFUND_BY_ADMIN, "Remboursement par un administrateur de l'évènement"),
    )

    owner = models.ForeignKey(User, related_name="registrations", on_delete=models.SET(User.get_anonyme_user))
    payee = models.CharField(max_length=64, blank=True, null=True)
    ticket = models.ForeignKey(Ticket, related_name="registrations", on_delete=models.CASCADE)
    options = models.ManyToManyField(Option, related_name="options_registration", blank=True)
    info_state = models.IntegerField(choices=STATE_INFO, default=STATE_CANNOT_PAID)
    purchase_date = models.DateTimeField(auto_now_add=True)
    paid = models.BooleanField(default=False)
    payment_method = models.IntegerField(choices=PAYMENT_METHODS, default=PAYMENT_UNKNOW)
    payment_date = models.DateTimeField(null=True)
    conditions_accepted = models.BooleanField(null=True)
    opposed = models.BooleanField(default=False)
    reason_opposed = models.IntegerField(choices=OPPOSED_REASON, default=OPPOSED_UNKNOW)
    refund = models.BooleanField(default=False)
    used = models.BooleanField(default=False)
    used_date = models.DateTimeField(blank=True, null=True)
    qrcode_uuid = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)
    lydia_request_uuid = models.CharField(default=None, null=True, max_length=32, unique=True)
    lydia_request_id = models.CharField(default=None, null=True, max_length=32)

    def get_payment_method(self):
        return self.PAYMENT_METHODS_DICT.get(self.payment_method, "Inconnue")

    def get_real_payee(self):
        return self.payee or self.owner.get_full_name()

    def get_price_options(self):
        price = 0
        for o in self.options.all():
            price += o.price
        return price

    def get_real_amount(self):
        return self.ticket.price + self.get_price_options()

    def __str__(self):
        return "{} – {} — {}".format(self.ticket.event.name, self.ticket.name, self.owner)

    def get_ipns(self):
        return PayPalIPN.objects.filter(
            invoice=(settings.PAYPAL_INVOICE_PREFIX + str(self.id)), payment_status=ST_PP_COMPLETED
        ).exclude(flag=True)

    def get_all_ipns(self):
        return PayPalIPN.objects.filter(invoice=(settings.PAYPAL_INVOICE_PREFIX + str(self.id)))

    def get_state_info_sup(self):
        d = {
            "accepted": [],
            "pending": [],
            "rejected": [],
            "missing": [],
        }
        for info in Info_Sup.objects.filter(ticket=self.ticket):
            if info.received.filter(registration=self).exists():
                info_recv = info.received.get(registration=self).get_sup_type()
                if info_recv.state == Info_Sup_Recv.STATE_ACCEPTED:
                    d["accepted"].append(info_recv)
                elif info_recv.state == Info_Sup_Recv.STATE_PENDING:
                    d["pending"].append(info_recv)
                elif info_recv.state == Info_Sup_Recv.STATE_MISSING:
                    d["missing"].append(info_recv)
                else:
                    d["rejected"].append(info_recv)
            else:
                __, __, cls = info.get_sup_type_prefixe()
                info_recv = cls(reference=info, registration=self, state=Info_Sup_Recv.STATE_MISSING)
                info_recv.save()
                d["missing"].append(info_recv)
        return d

    def get_list_info_sup(self):
        list_info = []
        for __, infos in self.get_state_info_sup().items():
            list_info.extend(infos)
        list_info.sort(key=(lambda x: x.reference.pk))
        return list_info

    def is_free(self):
        total = self.options.aggregate(price_options=Sum("price"))["price_options"]
        if total is None:
            total = 0
        total += self.ticket.price
        return total == 0


class LogEntry(models.Model):
    LOG_UNKNOW = 0
    LOG_PAYPAL = 1
    LOG_PAYPAL_ERROR = 2
    LOG_MANUAL_PAYEMENT = 3
    LOG_OPPOSITION_ADMIN = 4
    LOG_ENTRANCE = 5
    LOG_MANAGER = 6
    LOG_DOCUMENT = 7
    LOG_EVENT_MAJOR_CHANGE = 8
    LOG_PRINT_BILLET = 9
    LOG_FORCE_LIMIT = 10
    LOG_EMAIL_BILLET = 11
    LOG_LYDIA = 12
    LOG_LYDIA_ERROR = 13
    LOG_REASON = (
        (LOG_UNKNOW, "Sans attribut"),
        (LOG_PAYPAL, "Paiement paypal reçu"),
        (LOG_PAYPAL_ERROR, "Erreur Paiement paypal"),
        (LOG_MANUAL_PAYEMENT, "Paiement manuel"),
        (LOG_OPPOSITION_ADMIN, "Action d'opposition admin"),
        (LOG_ENTRANCE, "Billet scanné"),
        (LOG_MANAGER, "Gestion des administrateurs"),
        (LOG_DOCUMENT, "Reception d'information complémentaire"),
        (LOG_EVENT_MAJOR_CHANGE, "Modification d'un paramètre important sur l'évènement"),
        (LOG_PRINT_BILLET, "Impression billet par l'interface admin"),
        (LOG_FORCE_LIMIT, "Force les limites d'un billet"),
        (LOG_EMAIL_BILLET, "Envoie du billet par email"),
        (LOG_LYDIA, "Paiement Lydia reçu"),
        (LOG_LYDIA_ERROR, "Erreur paiement Lydia"),
    )

    date = models.DateTimeField(auto_now_add=True)
    user = models.ForeignKey(User, related_name="logs", on_delete=models.SET(User.get_anonyme_user))
    event = models.ForeignKey(Event, related_name="logs", null=True, on_delete=models.CASCADE)
    registration = models.ForeignKey(Registration, related_name="logs", null=True, on_delete=models.SET_NULL)
    id_registration = models.IntegerField(null=True, default=None)
    reason = models.IntegerField(choices=LOG_REASON, default=LOG_UNKNOW)
    operation = models.TextField()

    def __str__(self):
        return "{} {}: {}".format(self.date, self.user, self.operation)

    def get_operation(self):
        args = {}
        if self.registration:
            url = reverse("registration-edit", args=[self.registration.ticket.event.short_url, self.registration.id])
            args["registration"] = mark_safe('<a href="{}">#{}</a>'.format(url, self.registration.id))
        elif self.id_registration:
            args["registration"] = escape("#{}".format(self.id_registration))

        return mark_safe(escape(self.operation).format(**args))

    def save(self, *args, **kwargs):
        if self.id:
            raise SuspiciousOperation("On ne change pas les logs")
        if self.registration:
            self.id_registration = self.registration.id
        super(LogEntry, self).save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        raise SuspiciousOperation("On ne supprime pas les logs")


class Info_Sup(models.Model):
    """
    Un document complémentaire nécessaire à la validité du billet
    Un instance de Info_Sup_Recv décrit les infos reçus
    """

    INFO_SIMPLE = 0
    INFO_SHORT_TEXT = 1
    INFO_LONG_TEXT = 2
    INFO_CHOICE = 3
    INFO_INT = 4
    INFO_FLOAT = 5
    INFO_FILE = 6
    INFO_TYPE = (
        (INFO_SIMPLE, "Case à cocher dans l'interface d'administration"),
        (INFO_SHORT_TEXT, "Champ texte sur une ligne (max 128 caractères)"),
        (INFO_LONG_TEXT, "Champ texte multi-lignes"),
        (INFO_CHOICE, "Choix dans une liste"),
        (INFO_INT, "Nombre entier"),
        (INFO_FLOAT, "Nombre réel"),
        (INFO_FILE, "Fichier"),
    )
    INFO_DESC = {
        INFO_SIMPLE: "Simple",
        INFO_SHORT_TEXT: "Texte court",
        INFO_LONG_TEXT: "Texte long",
        INFO_CHOICE: "Choix",
        INFO_INT: "Nombre entier",
        INFO_FLOAT: "Nombre réel",
        INFO_FILE: "Fichier",
    }

    ticket = models.ForeignKey(Ticket, related_name="infos_sup", on_delete=models.CASCADE)
    name = models.CharField(max_length=32, verbose_name="Nom")
    description = models.TextField(
        blank=True, verbose_name="Description du document à fournir et information de transmision"
    )
    after_payement = models.BooleanField(
        default=False,
        verbose_name="Peut être validé après le paiement de la réservation "
        "(le billet ne sera imprimable qu'après validation)",
    )
    type_info = models.IntegerField(choices=INFO_TYPE, default=INFO_SIMPLE)
    info_validate = models.BooleanField(
        default=False,
        verbose_name="Auto-validation du champ dès qu'il est saisi. "
        "Si des informations sont en attentes au moment de l'activation, "
        "elles seront automatiquement validées.",
    )
    info_editable = models.BooleanField(
        default=False, verbose_name="Information modifiable si elle est validée " "(uniquement si auto-validation)"
    )

    def __str__(self):
        return "{} - {}".format(self.ticket.name, self.name)

    def get_sup_type_prefixe(self):
        type_ref = {
            self.INFO_SIMPLE: (None, None, Info_Sup_Recv),
            self.INFO_SHORT_TEXT: (None, "info_sup_recv_short_text", Info_Sup_Recv_Short_Text),
            self.INFO_LONG_TEXT: (None, "info_sup_recv_long_text", Info_Sup_Recv_Long_Text),
            self.INFO_CHOICE: ("info_sup_choice", "info_sup_recv_choice", Info_Sup_Recv_Choice),
            self.INFO_INT: ("info_sup_int", "info_sup_recv_int", Info_Sup_Recv_Int),
            self.INFO_FLOAT: ("info_sup_float", "info_sup_recv_float", Info_Sup_Recv_Float),
            self.INFO_FILE: ("info_sup_file", "info_sup_recv_file", Info_Sup_Recv_File),
        }
        if self.type_info in type_ref:
            return type_ref[self.type_info]
        return (None, None, Info_Sup_Recv)

    def get_sup_type(self):
        if isinstance(self, Info_Sup):
            prefix_info, __, __ = self.get_sup_type_prefixe()
            if prefix_info is not None:
                return getattr(self, prefix_info)
            else:
                return self
        return self

    def get_validators(self):
        if isinstance(self, Info_Sup) and self.get_sup_type() != self:
            return self.get_sup_type().get_validators()
        return []

    class Meta:
        verbose_name = "Information supplémentaire (tous)"
        verbose_name_plural = "Informations supplémentaires (tous)"


class ChoiceValidator(object):
    def __init__(self, valid_choice):
        self.valid_choice = valid_choice

    def __call__(self, value):
        if value not in self.valid_choice:
            raise ValidationError("Choix invalide", "invalid_choice")


class Info_Sup_Choice(Info_Sup):
    info_sup_ptr = models.OneToOneField(
        Info_Sup, on_delete=models.CASCADE, parent_link=True, related_name="info_sup_choice"
    )

    def get_validators(self):
        choices = self.choices.all()
        return [ChoiceValidator(choices)]

    class Meta:
        verbose_name = "Information supplémentaire (Choix)"
        verbose_name_plural = "Informations supplémentaires (Choix)"


class Info_Sup_Choice_Choices(models.Model):
    info_sup = models.ForeignKey(Info_Sup_Choice, related_name="choices", on_delete=models.CASCADE)
    name = models.CharField(max_length=32, verbose_name="Choix")

    class Meta:
        # On a pas deux fois le même noms
        unique_together = ["info_sup", "name"]

    def __str__(self):
        return "{} - {}".format(self.info_sup, self.name)


class Info_Sup_Int(Info_Sup):
    info_sup_ptr = models.OneToOneField(
        Info_Sup, on_delete=models.CASCADE, parent_link=True, related_name="info_sup_int"
    )
    min_value = models.IntegerField(
        blank=True, null=True, verbose_name="Valeur minimale (laissé vide si pas de minimum)"
    )
    max_value = models.IntegerField(
        blank=True, null=True, verbose_name="Valeur maximale (laissé vide si pas de maximum)"
    )

    def get_validators(self):
        validators = []
        if self.min_value is not None:
            validators.append(MinValueValidator(self.min_value))
        if self.max_value is not None:
            validators.append(MaxValueValidator(self.max_value))
        return validators

    class Meta:
        verbose_name = "Information supplémentaire (Entier)"
        verbose_name_plural = "Informations supplémentaires (Entier)"


class Info_Sup_Float(Info_Sup):
    info_sup_ptr = models.OneToOneField(
        Info_Sup, on_delete=models.CASCADE, parent_link=True, related_name="info_sup_float"
    )
    min_value = models.FloatField(blank=True, null=True, verbose_name="Valeur minimale (laissé vide si pas de minimum)")
    max_value = models.FloatField(blank=True, null=True, verbose_name="Valeur maximale (laissé vide si pas de maximum)")

    def get_validators(self):
        validators = []
        if self.min_value is not None:
            validators.append(MinValueValidator(self.min_value))
        if self.max_value is not None:
            validators.append(MaxValueValidator(self.max_value))
        return validators

    class Meta:
        verbose_name = "Information supplémentaire (Réel)"
        verbose_name_plural = "Informations supplémentaires (Réel)"


class Info_Sup_File(Info_Sup):
    info_sup_ptr = models.OneToOneField(
        Info_Sup, on_delete=models.CASCADE, parent_link=True, related_name="info_sup_file"
    )
    file_name = models.CharField(max_length=32, verbose_name="Nom du fichier (nom interne)", validators=[validate_slug])

    def get_validators(self):
        ext = ["pdf", "jpg", "jpeg", "png"]
        return [FileExtensionValidator(ext), FileValidator(), FileSizeValidator()]

    class Meta:
        verbose_name = "Information supplémentaire (Fichier)"
        verbose_name_plural = "Informations supplémentaires (Fichier)"


class Info_Sup_Recv(models.Model):
    """
    Un document complémentaire nécessaire à la validité du billet
    Si l'objet Registration contient une référence, l'utilisateur a fourni le document
    """

    STATE_ACCEPTED = 0
    STATE_PENDING = 1
    STATE_REJECTED = 2
    STATE_MISSING = 3
    STATE_STATUS = (
        (STATE_ACCEPTED, "Accepté"),
        (STATE_PENDING, "En attente"),
        (STATE_REJECTED, "Rejeté"),
        (STATE_MISSING, "Manquante"),
    )
    reference = models.ForeignKey(Info_Sup, related_name="received", on_delete=models.CASCADE)
    registration = models.ForeignKey(Registration, related_name="info_sup_received", on_delete=models.CASCADE)
    state = models.IntegerField(choices=STATE_STATUS, default=STATE_MISSING)
    # version must be update (regenerate) at each value update to avoid raced_condition
    version = models.UUIDField(default=uuid.uuid4, editable=True)

    def __str__(self):
        return "{} - {}".format(self.reference, self.registration)

    class Meta:
        unique_together = ["reference", "registration"]
        verbose_name = "Information supplémentaire reçue (tous)"
        verbose_name_plural = "Informations supplémentaires reçues (tous)"

    def update_version(self):
        # update only one time version per save
        old = Info_Sup_Recv.objects.get(pk=self.pk).version
        while old == self.version:
            self.version = uuid.uuid4()

    def get_sup_type(self):
        if isinstance(self, Info_Sup_Recv):
            __, prefix_info, __ = self.reference.get_sup_type_prefixe()
            if prefix_info is not None:
                return getattr(self, prefix_info)
            else:
                return self
        return self

    @property
    def value(self):
        if self == self.get_sup_type():
            return None
        return self.get_sup_type().value

    @value.setter
    def value(self, value):
        # must use sup_type
        # if set with use get_sup_type and not save, the value was lost
        raise ValueError("No settable")


class Info_Sup_Recv_Short_Text(Info_Sup_Recv):
    info_sup_recv_ptr = models.OneToOneField(
        Info_Sup_Recv, on_delete=models.CASCADE, parent_link=True, related_name="info_sup_recv_short_text"
    )
    text = models.CharField(max_length=128, blank=True)

    @property
    def value(self):
        return self.text

    @value.setter
    def value(self, value):
        self.text = value
        self.update_version()

    class Meta:
        verbose_name = "Information supplémentaire reçue (Texte court)"
        verbose_name_plural = "Informations supplémentaires reçues (Texte court)"


class Info_Sup_Recv_Long_Text(Info_Sup_Recv):
    info_sup_recv_ptr = models.OneToOneField(
        Info_Sup_Recv, on_delete=models.CASCADE, parent_link=True, related_name="info_sup_recv_long_text"
    )
    text = models.TextField(blank=True)

    @property
    def value(self):
        return self.text

    @value.setter
    def value(self, value):
        self.text = value
        self.update_version()

    class Meta:
        verbose_name = "Information supplémentaire reçue (Texte Long)"
        verbose_name_plural = "Informations supplémentaires reçues (Texte Long)"


class Info_Sup_Recv_Choice(Info_Sup_Recv):
    info_sup_recv_ptr = models.OneToOneField(
        Info_Sup_Recv, on_delete=models.CASCADE, parent_link=True, related_name="info_sup_recv_choice"
    )
    choice = models.ForeignKey(Info_Sup_Choice_Choices, on_delete=models.SET_NULL, null=True)

    @property
    def value(self):
        return self.choice

    @value.setter
    def value(self, value):
        self.choice = value
        self.update_version()

    class Meta:
        verbose_name = "Information Supplémentaire Reçue (Choix)"
        verbose_name_plural = "Informations supplémentaires reçues (Choix)"


class Info_Sup_Recv_Int(Info_Sup_Recv):
    info_sup_recv_ptr = models.OneToOneField(
        Info_Sup_Recv, on_delete=models.CASCADE, parent_link=True, related_name="info_sup_recv_int"
    )
    raw_value = models.IntegerField(null=True)

    @property
    def value(self):
        return self.raw_value

    @value.setter
    def value(self, value):
        self.raw_value = value
        self.update_version()

    class Meta:
        verbose_name = "Information Supplémentaire Reçue (Entier)"
        verbose_name_plural = "Informations supplémentaires reçues (Entier)"


class Info_Sup_Recv_Float(Info_Sup_Recv):
    info_sup_recv_ptr = models.OneToOneField(
        Info_Sup_Recv, on_delete=models.CASCADE, parent_link=True, related_name="info_sup_recv_float"
    )
    raw_value = models.FloatField(null=True)

    @property
    def value(self):
        return self.raw_value

    @value.setter
    def value(self, value):
        self.raw_value = value
        self.update_version()

    class Meta:
        verbose_name = "Information Supplémentaire Reçue (Réel)"
        verbose_name_plural = "Informations supplémentaires reçues (Réel)"


def path_info_recv(instance, filename):
    filename, file_extension = os.path.splitext(filename)
    random_str = get_random_string(length=12)
    info = instance.reference.get_sup_type()
    assert info.type_info == Info_Sup.INFO_FILE
    return "protected/billetterie/event_{0}/billet_{1}/{2}_{3}{4}".format(
        info.ticket.event.pk, instance.registration.id, info.file_name, random_str, file_extension
    )


class Info_Sup_Recv_File(Info_Sup_Recv):
    info_sup_recv_ptr = models.OneToOneField(
        Info_Sup_Recv, on_delete=models.CASCADE, parent_link=True, related_name="info_sup_recv_file"
    )
    upload_file = models.FileField(
        upload_to=path_info_recv, null=True, blank=True, verbose_name="Document fourni", max_length=128
    )

    @property
    def value(self):
        return self.upload_file

    @value.setter
    def value(self, value):
        if value is False:
            value = None
        self.upload_file = value
        self.update_version()

    class Meta:
        verbose_name = "Information Supplémentaire Reçue (Fichier)"
        verbose_name_plural = "Informations supplémentaires reçues (Fichier)"
